package main

import (
	"bufio"
	"encoding/csv"
	"fmt"
	"log"
	"math"
	"os"
	"strconv"
	"sync"
	"time"
)

func culMAE(records [][]string, relIndex int, predIndex int) float64 {
	length := float64(len(records))
	ret := 0.0
	for _, value := range records {
		a, _ := strconv.ParseFloat(value[relIndex], 64)
		b, _ := strconv.ParseFloat(value[predIndex], 64)
		ret += math.Abs(a - b)
	}
	ret /= length
	return ret
}
func culMSE(records [][]string, relIndex int, predIndex int) float64 {
	length := float64(len(records))
	ret := 0.0
	for _, value := range records {
		a, _ := strconv.ParseFloat(value[relIndex], 64)
		b, _ := strconv.ParseFloat(value[predIndex], 64)
		ret += math.Pow(a-b, 2.0)
	}
	ret /= length
	return ret
}
func culRMSE(records [][]string, relIndex int, predIndex int) float64 {
	length := float64(len(records))
	ret := 0.0
	for _, value := range records {
		a, _ := strconv.ParseFloat(value[relIndex], 64)
		b, _ := strconv.ParseFloat(value[predIndex], 64)
		ret += math.Pow(a-b, 2.0)
	}
	ret /= length
	return math.Sqrt(ret)
}

func culMAPE(records [][]string, relIndex int, predIndex int) float64 {
	length := float64(len(records))
	ret := 0.0
	for _, value := range records {
		a, _ := strconv.ParseFloat(value[relIndex], 64)
		b, _ := strconv.ParseFloat(value[predIndex], 64)
		ret += math.Abs((a - b) / a)
	}
	ret /= length
	return ret
}

// Diff cul diff
// input form:
// timestamp1 data1
// timestamp2 data2
// timestamp3 data3
//
// result:
// timestamp2 data2 - data1
// timestamp3 data3 - data2
func Diff(records [][]string) [][]string {
	var ret [][]string
	for i := 1; i < len(records); i++ {
		tempA, _ := strconv.ParseFloat(records[i][1], 32)
		tempB, _ := strconv.ParseFloat(records[i-1][1], 32)
		ret = append(ret, []string{records[i][0], strconv.FormatFloat(tempA-tempB, 'f', 5, 32)})
	}
	return ret
}

// ReadWaterData read water file
func ReadWaterData(waterFile string) [][]string {
	file, _ := os.Open(waterFile)
	r := csv.NewReader(file)
	records, err := r.ReadAll()
	if err != nil {
		log.Fatal(err)
	}
	return records
}

// WriteWaterData 写入水质数据
func WriteWaterData(waterFile string, resultData [][]string) {
	//print("resultData Size:", len(resultData))
	f, err := os.Create(waterFile)
	if err != nil {
		fmt.Printf("create map file error: %v\n", err)
		panic(err)
	}

	w := bufio.NewWriter(f)
	_, _ = fmt.Fprint(w, "Timestamp", ",", "real", ",", "predict\n")
	for _, line := range resultData {

		for lineI, l := range line {
			if lineI == 0 {
				_, _ = fmt.Fprint(w, l)
			} else {
				_, _ = fmt.Fprint(w, ",", l)
			}
		}
		_, _ = fmt.Fprint(w, "\n")
	}
	_ = w.Flush()
}

// Create a model where each possible y is the key of the map, and the set of all attributes of this y is the value of the map.
// modeSum is defined as the number of attribute entries added to the mode object. Its size = len(trainDiffSize) - L
func createMode(diff [][]string, length int) (map[string][][]string, int) {
	ret := make(map[string][][]string)
	var temp []string
	modeSum := 0
	for i := length; i < len(diff); i++ {
		temp = []string{}
		for j := i - length; j < i; j++ {
			temp = append(temp, diff[j][1])
		}
		modeSum++
		ret[diff[i][1]] = append(ret[diff[i][1]], temp)
	}
	return ret, modeSum
}

var quit chan int = make(chan int)
var FileLock = sync.Mutex{}
var threadSum = 0
var startLock = sync.Mutex{}
var endLock = sync.Mutex{}

func main() {
	mainFunc(5, 20)
}

var dataFileLock = sync.Mutex{}


func mainFunc(L int, m int) {
	startLock.Lock()
	threadSum++
	startLock.Unlock()

	// read water data
	waterFile := "luhaifeng_20160101_20200701.csv"
	waterRecord := ReadWaterData(waterFile)

	// Water quality data filtering, retaining only dissolved oxygen values and time stamps
	rjyRecord := ScanWaterRecord(waterRecord)
	dataSize := len(rjyRecord)

	trainDataSize := int(math.Ceil(float64(len(rjyRecord)) * 0.9))
	testDataSize := int(math.Floor(float64(len(rjyRecord)) * 0.1))


	// Delineate the data set
	trainData := rjyRecord[0:trainDataSize]
	testData := rjyRecord[dataSize-testDataSize:]

	dataFileLock.Lock()
	saveTrainDataAndTestData2File("trainData.csv", trainData, "testData.csv", testData)
	dataFileLock.Unlock()

	//  Judgment Training set size + Test set size = Total data size
	if len(trainData)+len(testData) != dataSize {
		panic("trainDataSize +testDataSize != dataSize")
	}

	trainDataDiff := Diff(trainData)
	testDataDiff := Diff(testData)

	mode, modeSum := createMode(trainDataDiff, L)
	D := culD(mode)
	// Making predictions about the test set
	result := predict(testDataDiff, mode, L, modeSum, D, m)

	// Restore to true value
	realResult := CulReal(testData, result)

	FileLock.Lock()
	// Write prediction results to file
	println("write to file!")
	WriteWaterData("result_L_"+strconv.Itoa(L)+"_m_"+strconv.Itoa(m)+".txt", realResult)
	// Calculate the error and write it to a file
	writeErrorValue("ErrorValue.txt", realResult, L, m)
	FileLock.Unlock()

	endLock.Lock()
	threadSum--
	endLock.Unlock()

	quit <- 0
}

// Determine if the given path file/folder exists
func Exists(path string) bool {
	_, err := os.Stat(path)
	return err == nil
}

func saveTrainDataAndTestData2File(trainDataFileName string, trainData [][]string, testDataFileName string, testData [][]string) {
	if !Exists(trainDataFileName) {
		WriteWaterData(trainDataFileName, trainData)
	}
	if !Exists(testDataFileName) {
		WriteWaterData(testDataFileName, testData)
	}
}

func CulReal(testData [][]string, result [][]string) [][]string {
	var trueResult [][]string
	for _, value := range result {
		timestamp := value[0]
		predictDiff, _ := strconv.ParseFloat(value[2], 32)
		find := 0
		for testDataI, testDataValue := range testData {
			if testDataValue[0] == timestamp {

				realPass, _ := strconv.ParseFloat(testData[testDataI-1][1], 32)
				PredictValue := realPass + predictDiff
				realValue, _ := strconv.ParseFloat(testData[testDataI][1], 32)
				trueResult = append(trueResult, []string{timestamp, strconv.FormatFloat(realValue, 'f', 5, 32), strconv.FormatFloat(PredictValue, 'f', 5, 32)})
				find = 1
				break
			}
		}
		if find == 0 {
			println(value[0], value[1])
			panic("Not fond in cul real Result!")
		}

	}

	return trueResult
}
// culD calculates the occurrence of each value at each position, and then requires D_x_i >= m before continuing
func culD(mode map[string][][]string) map[int]map[string]int {
	D := map[int]map[string]int{}
	for _, values := range mode {
		for _, value := range values {
			for i, xI := range value {
				if D[i] == nil {
					D[i] = map[string]int{}
					D[i][xI] = 0
				}
				D[i][xI]++
			}
		}
	}
	return D
}

func writeErrorValue(errorValueFile string, result [][]string, l int, m int) {
	f, _ := os.OpenFile(errorValueFile, os.O_RDWR|os.O_CREATE|os.O_APPEND, 0644)

	w := bufio.NewWriter(f)
	_, _ = fmt.Fprint(w, "l,", l, ",")
	_, _ = fmt.Fprint(w, "m,", m, ",")
	_, _ = fmt.Fprint(w, "RMSE,", culRMSE(result, 1, 2), ",")
	_, _ = fmt.Fprint(w, "MAPE,", culMAPE(result, 1, 2), ",")
	_, _ = fmt.Fprint(w, "MAE,", culMAE(result, 1, 2), ",")
	_, _ = fmt.Fprint(w, "MSE,", culMSE(result, 1, 2), "\n")
	_ = w.Flush()
}

// predict predict function
func predict(testDiff [][]string, mode map[string][][]string, L int, modeSum int, D map[int]map[string]int, m int) [][]string {
	if L >= len(testDiff) {
		panic("L >= len(testDiff)")
	}
	var ret [][]string

	// Property set x
	var x []string
	for i := 0; i < L; i++ {
		x = append(x, testDiff[i][1])
	}

	startTime := time.Now()

	// start prediction
	for index := L; index < len(testDiff); index++ {
		if (index-L)%100 == 0 {
			elapsedTime := time.Since(startTime)
			fmt.Println("L=", L, "m=", m, "-", index-L, "/", len(testDiff)-L, "Spent time:", elapsedTime)
			startTime = time.Now()
		}

		maxProbability := 0.0
		predictValue := "0.00000"

		measuredValue := testDiff[index][1]
		timestamp := testDiff[index][0]

		// the probability of each set of y values
		// Calculate the probability of each type of y
		for possibleY := range mode {
			if float64(len(mode[possibleY])) < float64(modeSum)*float64(m)/10000 {
				continue
			}

			innerValue1 := 0.0
			for i := 0; i < L; i++ {

				// Here the threshold constant needs to be added later to speed up the calculation
				P_c_x_i, D_c_x_i := cul_P_c_x_i(mode, possibleY, i, x, modeSum)

				innerValue2 := 0.0
				for j := 0; j < L; j++ {
					P_x_j_c_x_i := cul_P_x_j_c_x_i(mode, possibleY, i, j, x, D_c_x_i)
					if innerValue2 == 0.0 {
						innerValue2 = P_x_j_c_x_i
					} else {
						innerValue2 *= P_x_j_c_x_i
					}
				}
				innerValue1 += P_c_x_i * innerValue2
			}

			if innerValue1 > maxProbability {
				maxProbability = innerValue1
				predictValue = possibleY
			}
		}

		ret = append(ret, []string{timestamp, measuredValue, predictValue})


		// then generate the next set of X , thus predicting the next value
		x = x[1:]
		x = append(x, testDiff[index][1])
	}
	return ret
}

func cul_P_x_j_c_x_i(mode map[string][][]string, y string, i int, j int, x []string, D_c_x_i int) float64 {
	D_c_x_i_x_j := 0
	for _, value := range mode[y] {
		if value[i] == x[i] && value[j] == x[j] {
			D_c_x_i_x_j += 1
		}
	}

	// The number of possible values for the i-th attribute in mode[y]
	Nj := 0
	var tempNj map[string]int
	for _, list := range mode[y] {
		if tempNj[list[j]] != 0 {
			Nj++
			tempNj[list[j]]++
		}
		if tempNj[list[j]] != 1 && tempNj[list[j]] != 0 {
			panic("tempNj[list[j]] != 1 && tempNj[list[j]] != 0 ")
		}
	}
	return float64(D_c_x_i_x_j+1) / float64(D_c_x_i+Nj)
}

// cul_P_c_x_i computes P(c,x_i)
// mode is the set of possible y and corresponding attributes map
// y is the current prediction
// i is the subscript number of the current attribute x to be computed as i
// modeSum is the number of attribute entries added to the mode object
func cul_P_c_x_i(mode map[string][][]string, y string, i int, x []string, modeSum int) (float64, int) {
	D_c_x_i := 0

	for _, value := range mode[y] {
		if value[i] == x[i] {
			D_c_x_i += 1
		}
	}

	// D is the total number of samples
	D := float64(modeSum)

	// N is the number of label categories
	N := len(mode)

	// The number of possible values for the i-th attribute in mode[y]
	Ni := 0
	var tempNi map[string]int
	for _, list := range mode[y] {
		if tempNi[list[i]] != 0 {
			Ni++
			tempNi[list[i]]++
		}
		if tempNi[list[i]] != 1 && tempNi[list[i]] != 0 {
			panic("tempNi[list[i]] != 1 && tempNi[list[i]] != 1")
		}
	}

	return (float64(D_c_x_i) + 1) / (D + float64(N*Ni)), D_c_x_i
}

// ScanWaterRecord extracts timestamps and dissolved oxygen values from water quality datasets
func ScanWaterRecord(record [][]string) [][]string {
	var rjyRecord [][]string
	rjyIndex := -1
	timestampIndex := -1
	for ix, recordValue := range record {
		if ix == 0 {
			for itemIndex, item := range recordValue {
				if item == "Timestamp" {
					timestampIndex = itemIndex
				} else if item == "rjy_con" {
					rjyIndex = itemIndex
				}
			}
		} else {
			if recordValue[rjyIndex] == "-1" {
				continue
			}
			rjyRecord = append(rjyRecord, []string{recordValue[timestampIndex], recordValue[rjyIndex]})
		}
	}
	return rjyRecord
}